#!/usr/bin/python
# -*- coding: utf-8 -*-

import io, zipfile, sys, re, subprocess, os, tempfile
import html2text, odt2txt
import xml.etree.ElementTree as ET

try:
    # Python3
    from urllib.request import urlopen
except:
    # Python 2
    from urllib2 import urlopen

try:
    # Nur unter Windows
    from win32com.client import Dispatch
except:
    pass

#################################################################
# Verwendung:
# Klasseninstanz erzeugen (mit Dateinamen und Kodierung)
# Data sind über Klasseninstanz.txt erreichbar
#################################################################

class ConvertText:
    def __init__(self, filename='', encoding='utf-8'):
        """Textformat bestimmen und dann entsprechende Funktion aufrufen"""
        self.filename = filename
        self.encoding = encoding
        
        if self.filename.endswith(('.txt', '.md')):
            self.txt = self.read_txt()
            
        elif self.filename.endswith('.docx'):
            self.txt = self.read_docx()

        elif self.filename.endswith('.doc'):
            self.txt = self.read_doc()

        elif self.filename.endswith('.odt'):
            self.txt = self.read_odt()

        elif self.filename.startswith('http'):
            self.txt = self.read_html()

        elif self.filename.endswith(('.html', '.htm')):
            self.txt = self.read_html()
            
        elif self.filename.endswith('.pdf'):
            self.txt = self.read_pdf()
            
        else:
            # wenn nichts geht
            self.txt = ''
        


    def read_txt(self):
        """Platte Textfiles lesen"""
        # io.open (statt einfach open) funktioniert auch für Python 2.7
        f = io.open(self.filename, encoding=self.encoding)
        txt = f.read()
        return txt

            
    def read_odt(self):
        """OpenOffice-Files lesen"""
        txt = odt2txt.OpenDocumentTextFile(self.filename)
        txt = txt.toString()                # gibt Unicode aus
        return txt


    def read_docx(self):
        """DOCX Word-Files lesen"""
        """Vgl.: http://etienned.github.io/posts/extract-text-from-word-docx-simply/ """
        
        WORD_NAMESPACE = '{http://schemas.openxmlformats.org/wordprocessingml/2006/main}'
        PARA = WORD_NAMESPACE + 'p'
        TEXT = WORD_NAMESPACE + 't'
        
        with zipfile.ZipFile(self.filename, "r") as zf:
            inhalte = []
            for doc in ['word/document.xml', 'word/footnotes.xml', 'word/endnotes.xml']:
                try:
                    xml_content = zf.read(doc)
                    tree = ET.XML(xml_content)
                     
                    for paragraph in tree.getiterator(PARA):
                        texts = [node.text
                                for node in paragraph.getiterator(TEXT)
                                if node.text]
                        if texts:
                            res = ''.join(texts)
                            # res = ' '.join(res.split())
                            inhalte.append(res)
                except:
                    pass

            txt = '\n'.join(inhalte)
        return txt


    def read_doc(self):
        """Alte DOC Word-Files lesen"""
        w = Dispatch("Word.Application")
        # w.Visible = 1
        w.Documents.Open(self.filename)
        txt = w.ActiveDocument.Content.Text
        txt = ' '.join(txt.split())
        w.ActiveDocument.Close()
        return txt


    def read_html(self):
        """HTML parsen"""
        # Datei aus dem Netz holen
        if self.filename.startswith('http'):
            f = urlopen(self.filename)
            txt = f.read()
            # Versuchen, die Kodierung aus dem Header zu holen
            try:
                encoding = f.headers['content-type'].split('charset=')[-1]
                txt = txt.decode(encoding)
            # Probieren wir, die Kodierung aus dem File auszulesen
            except LookupError:
                temp = str(txt)
                encoding = re.search("text/html; charset=([ A-Za-z0-9-]+)", temp)
                if encoding != None:
                    encoding = encoding.group(1).strip()
                    txt = txt.decode(encoding)
            
        # oder aber lokale Datei einlesen
        else:
            # io.open (statt einfach open) funktioniert auch für Python 2.7
            f = io.open(self.filename, encoding=self.encoding)
            txt = f.read()

        # Konvertierung mit html2text zu Markdown
        # https://pypi.python.org/pypi/html2text
        h = html2text.HTML2Text()
        h.ignore_links = 1
        h.ignore_images = 1
        h.ignore_emphasis = 1
        txt = h.handle(txt)

        return txt


    def read_pdf(self):
        # Funktioniert nur unter Linux
        try:
            tmp_file = os.path.join(tempfile.gettempdir(), 'TextSTAT_tmp.txt')
            subprocess.call(["pdftotext", "-nopgbrk", self.filename, tmp_file])
            self.filename = tmp_file
            txt = self.read_txt()
            os.remove(tmp_file)
            return txt
        except:
            return ''
            

if __name__=="__main__":
    f = "test.docx"
    f = 'http://stern.de'
    f = "test.pdf"

    t = ConvertText(f)
    print(t.txt.encode('utf-8'))
        
