#!/usr/bin/python
# -*- coding: utf-8 -*-

import os.path, pickle, sqlite3, re
import ConvertText
try:
    # Python3
    from urllib.request import urlopen
    from urllib.parse import urljoin
except:
    # Python 2
    from urllib2 import urlopen
    from urlparse import urljoin

import locale
locale.setlocale(locale.LC_CTYPE, '')     # LC_CTYPE beeinflusst String-Operationen
locale.setlocale(locale.LC_COLLATE, '')   # LC_COLLATE beeinflusst das Sortier-Verhalten


class Korpus:
    def __init__(self, master=None):
        self.master = master
        self.dbFile = None
        self.dbName = None
        self.db = None          # dies ist die SQLite Datenbank Connection

        self.freqListe = []
        self.concListe = []        


    def clear_results(self):
        self.freqListe = []
        self.concListe = []
        self.zitatDateiPfad = ''


    def db_connect(self, f):
        """Datenbankverbindung herstellen"""
        self.clear_results()
        try:
            self.dbFile = f
            self.dbName = os.path.split(self.dbFile)[1]
            # Datenbank öffnen
            self.db = sqlite3.connect(self.dbFile)
            # Connection mit Spaltennamen
            self.db.row_factory = sqlite3.Row
            # Locale zum Sortieren
            self.db.create_collation("lokal", locale.strcoll)
            # SQLite Journal-Files nicht auf Disk schreiben
            self.db.execute('PRAGMA journal_mode=MEMORY;')

            # Funktion zum retrograden Sortieren
            def collate_retro(a, b):
                a, b = list(a), list(b)
                a.reverse()
                b.reverse()
                a = ''.join(a)
                b = ''.join(b)
                return locale.strcoll(a,b)
    
            self.db.create_collation("retro", collate_retro)
            
        except:
            print ('Database error!')
            self.dbFile = None
            self.dbName = None
            self.db = None


    def db_close(self):
        """Datenbank schließen"""
        # erst Änderungen wegschreiben
        if self.db != None:
            self.db.commit()
            self.db.close()
            
        self.clear_results()
        self.db = None
        self.dbFile = None
        self.dbName = None


    def db_new(self, f):
        """Neue Datenbank anlegen"""
        self.db_connect(f)
        self.db.execute( "CREATE TABLE texte (id INTEGER PRIMARY KEY, pfad TEXT UNIQUE, txt TEXT, anz_tokens INTEGER, anz_types INTEGER)" )
        self.db.execute( "CREATE TABLE wortliste (id INTEGER PRIMARY KEY, wort TEXT, wortklein TEXT, lemma TEXT, pos TEXT, txt_id INTEGER)")
        self.db.commit()


    def db_open(self, f):
        """Datenbank öffnen""" 
        self.db_connect(f)


    def db_save(self):
        """Änderungen am Korpus in DB schreiben"""
        self.db.commit()


    def file_add(self, filename='', txt='', encoding='utf-8'):
        """Eine Datei zur Datenbank hinzufügen"""
        
        # Manchmal wird schon Text übergeben (beim Importieren eines (alten) Korpus)
        # Wenn aber kein Text übergeben wird, dann erst File lesen und konvertieren
        if txt == '':
            f = ConvertText.ConvertText(filename, encoding)
            # Text holen
            txt = f.txt
            # Typographische Anführungszeichen ersetzen
            txt = re.sub("[„“”«»]", '"', txt)
            txt = re.sub("[‚‘’‹›]", "'", txt)
            # Dashes ersetzen: ‒ – — ― (sehen hier aus wie *ein* langer Strich...
            txt = re.sub("[‒–—―]", '-', txt)
            
        # Wenn was schiefgegangen ist und kein Text vorhanden ist
        if txt == '':
            raise Exception("File not added to database. Text could not be read.")
       
        # Ohne Satzzeichen
        muster = re.compile(r"\b([\w·'-]+)\b", re.U | re.M)
        woerter = muster.findall(txt)

        # Mit Satzzeichen
        # Dies funktioniert wohl nur mit Python3 (und auch da nicht zuverlässig)
#        woerter = re.split("(\W)", txt, re.U)
#        woerter = [w for w in woerter if w not in ['', ' ', '\n']]
        
        anz_tokens = len(woerter)
        anz_types = len(set(woerter))

        # Text in Datenbank einfügen (jetzt noch ohne genre)
        try:
            cur = self.db.cursor()
            cur.execute("INSERT INTO texte(pfad, txt, anz_tokens, anz_types) VALUES(?,?,?,?)", (filename, txt, anz_tokens, anz_types) )
            txt_id = cur.lastrowid
        except sqlite3.IntegrityError:
            # da die Datei schon vorhanden ist, muss sie zunächst gelöscht werden
            self.file_del(filename=filename)
            # So, jetzt aber: einfügen
            cur = self.db.cursor()
            cur.execute("INSERT INTO texte(pfad, txt, anz_tokens, anz_types) VALUES(?,?,?,?)", (filename, txt, anz_tokens, anz_types) )
            txt_id = cur.lastrowid

        # Wortliste machen und einfügen (case-sensitive, case-insensitive, ohne lemma und pos)
        wortliste = [[w, w.lower(), txt_id] for w in woerter]
        self.db.executemany("INSERT INTO wortliste(wort, wortklein, txt_id) VALUES(?, ?, ?)", wortliste)

        self.clear_results()  
        

    def file_del(self, filename='', txt_id=0):
        """Eine Datei aus der Datenbank entfernen"""
        cur = self.db.cursor()
        
        # Wenn die ID des Eintrags nicht bekannt ist
        if filename != '' and txt_id == 0:
            cur.execute("SELECT id FROM texte WHERE pfad = ?", (filename,))
            res = cur.fetchone()
            txt_id = res['id']
            
        # Eintrag anhand der ID löschen
        cur.execute("DELETE FROM texte WHERE id = ?", (txt_id,))
        # auch die dazugehörigen Einträge aus der Wortliste löschen
        cur.execute("DELETE FROM wortliste WHERE txt_id = ?", (txt_id,))
        
        self.clear_results()         


    def file_get(self, filename=''):
        """Inhalt einer Datei aus der Datenbank holen"""
        cur = self.db.cursor()
        cur.execute("SELECT id, pfad, txt FROM texte WHERE pfad = ?", (filename,))
        res = cur.fetchone()
        return res['txt']


    def db_import(self, filename=''):
        """Korpus hinzufügen / importieren"""
        if os.path.exists(filename):
            # Neues TextSTAT 3 Korpus importieren
            try:
                con = sqlite3.connect(filename)
                
                groesse = con.execute("SELECT count(*) FROM texte")
                groesse = list(groesse)[0][0]
                # Progressbar einstellen
                self.master.status.progress.config(maximum = groesse) 
                
                for row in con.execute("SELECT pfad, txt FROM texte"):
                    self.file_add(filename=row[0], txt=row[1])
                    
                    # Fortschritt in Statusbar anzeigen
                    self.master.status.progress.step(1)
                    self.master.status.progress.update()
            
                con.close()

            except:
                # Vielleicht TextSTAT 2 Korpus?
                with open(filename, 'rb') as f:
                    db = pickle.load(f)
                    
                    # Progressbar einstellen
                    self.master.status.progress.config(maximum = len(db)) 
                    for k in db:
                        self.file_add(filename=db[k]['path'], txt=db[k]['file'])

                        # Fortschritt in Statusbar anzeigen
                        self.master.status.progress.step(1)
                        self.master.status.progress.update()
                        
            finally:
                # Änderungen in Datenbank schreiben
                self.db.commit()
                # Progressbar zurücksetzen
                self.master.status.progress.config(value = 0)
                self.master.status.progress.update()
                            

    def getDbName(self):
        """Gibt Dateinamen des Korpus zurück"""
        return os.path.split(self.dbFile)[1]

        
    def getDbFileNames(self):
        """Gibt sortierte Liste der Filenamen zurück (Unicode)"""
        l = []
        for row in self.db.execute("SELECT pfad FROM texte ORDER BY pfad"):
            l.append(row['pfad'])
        return l


    def getDbLength(self):
        """Gibt die Zahl er Text in der Datenbank zurück"""
        cur = self.db.cursor()
        anz = cur.execute("SELECT count(*) FROM texte")
        anz = cur.fetchone()
        return anz[0]


    def getDbSize(self):
        """Gibt die Anzahl der Tokens und der Types in der Datenbank aus"""
        cur = self.db.cursor()
        cur.execute("SELECT sum(anz_tokens), sum(anz_types) FROM texte")
        res = cur.fetchone()
        # Tuple (anz_tokens, anz_types)
        return res
    
    
    def getFileInfo(self, filename=''):
        cur = self.db.cursor()
        cur.execute("SELECT anz_tokens, anz_types FROM texte WHERE pfad=?", (filename,))
        res = cur.fetchone()
        # Tuple (anz_tokens, anz_types)
        return res


    def getWordFreq(self, sort='freq', grossklein=0, minFreq=0, maxFreq=0, suchFreq=''):
        """Gibt Wortfrequenzliste zurück"""
        
        try: minFreq = int(minFreq)
        except: minFreq = 0
        try: maxFreq = int(maxFreq)
        except: maxFreq = 0
        
        # Suchspalte abhängig von case-sensitive oder nicht
        if grossklein == 1:
            suchFreq = suchFreq.lower()
            sql = "SELECT wortklein AS w, count(*) as anz FROM wortliste"
        else:
            sql = "SELECT wort AS w, count(*) as anz FROM wortliste"
        
        # Checken ob Suchbegriff oder min./max. Frequenzen
        # Suchbegriff
        if suchFreq != '':
            sql = sql + " WHERE w GLOB '%s'" % suchFreq
               
        
        # Sortierung
        if sort == 'alpha':
            sql = sql + " GROUP BY w ORDER BY w COLLATE lokal"
        elif sort == 'retro':
            sql = sql + " GROUP BY w ORDER BY w COLLATE retro"
        else:
            sql = sql + " GROUP BY w ORDER BY count() DESC, w COLLATE lokal"

        
        # Jetzt in DB suchen
        res = self.db.execute(sql)

        # Frequenzbeschränkungen?
        if minFreq ==0 and maxFreq==0:
            self.freqListe = [ [x[0], x[1]] for x in res]
        else:
            self.freqListe = []
            for k, v in res:
                if (maxFreq == 0 and v >= minFreq) or (v >= minFreq and v <= maxFreq):
                    self.freqListe.append([k, v])

        return self.freqListe


    def getWordConc(self, such='', ganzWort=0, grossklein=0, cL='50', cR='50', sort='', markKonk=0):
        """Konkordanzen generieren"""       
        try:
            cL = int(cL)
            cR = int(cR)
        except:
            cL = 50
            cR = 50
            
        if ganzWort == 1:
            # nur ganze Wörter
            suchmuster = r"\b%s\b" % such
        else:
            suchmuster = such
            
        if grossklein == 1:
            # Großschreibung ignorieren
            muster = re.compile(suchmuster, re.U | re.S | re.I)
        else:
            muster = re.compile(suchmuster, re.U | re.S)

        # Progressbar einstellen
        self.master.status.progress.config(maximum = self.getDbLength() )
        # Liste leer machen
        self.concListe = []
        # Jetzt suchen in Datenbank
        for row in self.db.execute("SELECT id, pfad, txt FROM texte"):
            datei = row['txt']
            
            pos = 0
            while 1:
                res = muster.search(datei, pos)
                if res == None:
                    break
    
                # ab pos weitersuchen, also pos erhöhen
                pos = res.start() + 1
    
                # Kontext festlegen
                anfang = res.start() - cL
                ende = res.end() + cR

                if anfang < 0:
                    anfang = 0
                if ende > len(datei):
                    ende = len(datei)
    
                # Gefundenen String (besteht selber aus drei Teilen: Kontext links, Treffer, Kontext rechts) 
                # ins Resultat, dazu Pfad und Datei-ID
                konk = [datei[anfang:res.start()], datei[res.start():res.end()], datei[res.end():ende], row['pfad'], row['id']]
                self.concListe.append(konk)
                
            # Fortschritt in Statusbar anzeigen
            self.master.status.progress.step(1)
            self.master.status.progress.update()

        # Progressbar zurücksetzen
        self.master.status.progress.config(value = 0)
        self.master.status.progress.update()

        # Sortieren (jetzt NICHT locale-abhängig)
        if sort == 'rechts':
            self.concListe.sort(key=lambda i: i[2].lower())
        elif sort == 'links':
            self.concListe.sort(key=lambda i: i[0].split()[-1].lower())
        else:
            self.concListe.sort(key=lambda i: i[1].lower())

        return self.concListe


    def getContext(self, i, contextL=200, contextR=200):
        """Mehr Kontext suchen"""
        konk = self.concListe[i]
        konkordanz = ''.join(konk[0:3])
        txt_id = konk[4]
        
        cur = self.db.cursor()
        cur.execute("SELECT id, pfad, txt FROM texte WHERE id = ?", (txt_id,))
        res = cur.fetchone()

        self.zitatDateiPfad = res['pfad']
        datei = res['txt']
        
        stelleL = datei.find(konkordanz)
        stelleR = stelleL + len(konkordanz)
        
        anfang = stelleL - contextL
        ende = stelleR + contextR

        # Leerzeichen suchen, damit nur ganze Wörter gezeigt werden
        if anfang < 0: 
            anfang = 0
        while datei[anfang] != ' ':
            anfang = anfang - 1
            if anfang <= 0:
                anfang = 0
                break
        if ende > len(datei): 
            ende = len(datei)-1
        while datei[ende] != ' ':
            ende = ende + 1
            if ende >= len(datei):
                ende = len(datei)
                break

        res = [datei[anfang:stelleL], datei[stelleL:stelleR], datei[stelleR:ende]]
        return res


### Web-Crawler ###

    def crawlWeb(self, url='', anzahl=1, basis=''):
        """Crawle bestimmte Anzahl von Websiten für das Korpus"""
        
        # Suchbereich festlegen    
        if basis == 'server':
            basis = urljoin(url, '/')
        else:
            basis = urljoin(url, ' ').strip()

        url_liste = []
        url_liste.append(url)

        # Progressbar einstellen
        self.master.status.progress.config(maximum = anzahl )

        i = 0
        erfolg = 0
        while i < len(url_liste) and erfolg < anzahl:
            adresse = url_liste[i]
            # Zähler hochsetzen
            i = i+1
            # URL lesen und Links darin suchen
            try:
                f = urlopen(adresse)
                txt = f.read()
                # Versuchen, die Kodierung aus dem Header zu holen
                try:
                    encoding = f.headers['content-type'].split('charset=')[-1]
                    txt = txt.decode(encoding)
                # Wenn das schiefgeht, dann suchen wir die Kodierung im File
                except LookupError:
                    temp = str(txt)
                    encoding = re.search("text/html; charset=([ A-Za-z0-9-]+)", temp)
                    if encoding != None:
                        encoding = encoding.group(1).strip()
                        txt = txt.decode(encoding)

                # Links extrahieren
                urls = re.findall('''href=["'](.[^"']+)["']''', txt, re.I)
                for x in urls:
                    # falls es eine relative URL ist (ansonsten wird die komplette URL ersetzt)
                    x = urljoin(url, x)
                    if x.startswith(basis):
                        if not x.endswith(('.css', '.js', '.rss', '.pdf', '.doc', '.docx', '.gif', '.bmp', '.png', '.jpg', '.jpeg', '.ico', '.wav', '.mp3', '.mp4', '.mov', '.ogg')):
                            if '#' not in x:
                                if x not in url_liste:
                                    url_liste.append(x)
                                    
            # Wenn was schiefgegangen ist, dann einfach die nächste URL probieren
            except Exception as e:
                # print (e)
                continue

            # Wenn bis hierhin alles gut gegangen ist:
            # File ins Korpus
            self.file_add(filename=adresse)
            erfolg=erfolg + 1
            
            # Fortschritt in Statusbar anzeigen
            self.master.status.progress.step(1)
            self.master.status.progress.update()

 
        # Progressbar zurücksetzen
        self.master.status.progress.config(value = 0)
        self.master.status.progress.update()

